//Processing script for Forest Garden plan creation
//Simon Parker
//19/5/2011 - Updated to allow interactive drawing of plant locations and sizes

/**
configfiles taken from http://wiki.processing.org/index.php/External_configuration_files
@author toxi
**/

ArrayList plants;
ArrayList sfruit;
ArrayList drawnPlants;
int width = 1100;
int height = 550;
int frameTrig=0;
float scale = width/35.0;
float offset = 50;
float t_offs = 10;
float rot = 0.0; //not implemented correctly yet
float centerX;
float centerY;
float outerX;
float outerY;
PFont font;
boolean datesOn = false;
boolean rtsOn = false;
boolean editMode = false;
boolean imageOn = true;
boolean imageLoaded = false;
boolean showInstr = true;
PImage p;
P5Properties props;

void drawGrid() {
  stroke(128, 128, 128,128);
  for (int i=0; i<width/scale; i++) {
    line(i*scale, 0, i*scale, height);
  }
  for (int i=15; i>0; i--) {
    line(0, i*scale-offset, width, i*scale-offset);
  }
}
void drawTrees() {
  for (int i=0 ; i<plants.size() ; i++) {
    fill(0, 255, 0, 50);
    ((Plant) plants.get(i)).drawGlyph();
  }
}
void drawSFruit() {
  for (int i=0 ; i<sfruit.size() ; i++) {
    fill(0, 128, 128, 50);
    ((Plant) sfruit.get(i)).drawSFGlyph();
  }
}
void drawDPlants() {
  for (int i=0 ; i<drawnPlants.size() ; i++) {
    fill(0, 0, 255, 50);
    ((Plant) drawnPlants.get(i)).drawDPGlyph();
  }
}
void drawPlants() {
  drawTrees();
  drawSFruit();
  drawDPlants();
}

void drawSFNames() {
  for (int i=0 ; i<sfruit.size() ; i++) {
    ((Plant) sfruit.get(i)).drawLineText(datesOn);
  }
}
void drawTNames() {
  for (int i=0 ; i<plants.size() ; i++) {
    ((Plant) plants.get(i)).drawText(datesOn, rot);
  }
}
void drawLegend() {
  float xt = width *0.8;
  float yt = height * 0.1;
  float l_os = 25;
  for (int i=0 ; i<sfruit.size() ; i++) {
    yt = ((Plant) sfruit.get(i)).drawLegend(xt, yt, l_os) ;
    //xt = xt + l_os; 
    //yt = yt + l_os;
  }
}
void drawScale() {
//  stroke(0, 0, 0);
//  fill(0, 0, 0);
//  line(2.0*scale, height-1.0*scale-offset, 7.0*scale, height-1.0*scale-offset);
//  for (int i=0; i<6; i++) {
//    line((2.0+i)*scale, height-(1.0-0.2)*scale-offset, (2.0+i)*scale, height-1.0*scale-offset);
//    String s = (i + "m");
//    text(s, (1.9+i)*scale, height-(1.0-0.5)*scale-offset);
//  }
  //  //draw compass N
  //  line(20*scale,height-10*scale-offset,24*scale,height-10.2*scale-offset);
  //  line(24*scale,height-10.2*scale-offset,23.5*scale,height-10.4*scale-offset);
  //  line(24*scale,height-10.2*scale-offset,23.55*scale,height-9.95*scale-offset);
  //  text("N", (22)*scale, height-10.2*scale-offset);
}
void drawImage(){

  if (imageOn && imageLoaded) {
    image(p,0,0);
  }
  if (showInstr) {
      stroke(0,0,255);
      fill(255,255,255,128);
      rect(5,5,400,95);
      fill(0,0,255);
      text("e/E - edit mode (click, drag, release to draw)    v/V - view mode",10,20);
      text("w - write drawn plants to drawnplants.csv",10,35);
      text("x/X - hide background    b/B - show background",10,50);
      text("i/I - write image file to plan.png",10,65);
      text("p - draw all plants    t - draw trees only    s - draw soft fruit only",10,80);  
      text("h/H - hide instructions    j/J - show instructions",10,95);
    }
}


//void drawEditMode(boolean bEM) {
//  if (bEM) {
//    String em = ("Edit mode");
//    stroke(255, 0, 0);
//    fill(255, 0, 0);
//    text(em, (1.9+8)*scale, height-(1.0-0.5)*scale-offset);
//  }
//  else {
//    stroke(255, 255, 255);
//    fill(255, 255, 255);
//    rect((9.5)*scale, height-(1.0)*scale-offset, 3.5*scale, 0.8*scale);
//  }
//}

void readTrees() {
  String[] lines = loadStrings("settings/trees.csv");
  print("\nReading settings/trees.csv\n");
  for (int i=1 ; i<lines.length ; i++) {
    print(lines[i]+"\n");
    String[] data = split(lines[i], ',');
    float x = Float.valueOf(data[0]).floatValue();
    float y = Float.valueOf(data[1]).floatValue();
    float r = Float.valueOf(data[2]).floatValue();
    plants.add( new Plant( x, y, r, data[3], data[4], "-", true ) );
  }
}
void readDrawPlotBoundary() {
  stroke(0, 0, 0, 150);
  String[] lines2 = loadStrings("settings/boundary.csv");
  print("\nReading settings/boundary.csv\n");
  for (int i=1 ; i<lines2.length ; i++) {
    //print(lines2[i]+"\n");
    String[] data2 = split(lines2[i], ',');
    float x1 = Float.valueOf(data2[0]).floatValue();
    float y1 = Float.valueOf(data2[1]).floatValue();
    float x2 = Float.valueOf(data2[2]).floatValue();
    float y2 = Float.valueOf(data2[3]).floatValue();
    line(x1*scale, height-y1*scale-offset, x2*scale, height-y2*scale-offset);
  }
}
void readSoftFruit() {
  ArrayList names;
  names = new ArrayList();

  String[] lines3 = loadStrings("settings/softfruit.csv");
  print("\nReading settings/softfruit.csv\n");
  for (int i=1 ; i<lines3.length ; i++) {
    boolean LE = true;
    print(lines3[i]+"\n");
    String[] data3 = split(lines3[i], ',');
    float x3 = Float.valueOf(data3[0]).floatValue();
    float y3 = Float.valueOf(data3[1]).floatValue();
    float r3 = Float.valueOf(data3[2]).floatValue();
    if (names.contains(data3[5])) {
      LE = false;
    }
    names.add(data3[5]);
    sfruit.add( new Plant(x3, y3, r3, data3[3], data3[4], data3[5], LE));
  }
}  
void readData() {
  readTrees();
  readDrawPlotBoundary();
  readSoftFruit();
}
void writeDrawnPlants() {
  String[] strToSave = {"Copy the rows into either trees.csv or softfruit.csv and add names","x,y,r,plant,variety,acronym (for softfruit)"};
  for (int i=0 ; i<drawnPlants.size() ; i++) {
    String s = str(((Plant) drawnPlants.get(i)).x) +","+ str(((Plant) drawnPlants.get(i)).y) +","+ str(((Plant) drawnPlants.get(i)).d)+",tbc,tbc,tbc";
    strToSave = append(strToSave,s);
  }
  saveStrings("settings/drawnplants.csv", strToSave); 
  print("Written to drawnplants.csv");
}

void setup() {
  try {
    props=new P5Properties();
    props.load(openStream("settings/conf.properties"));
    scale = props.getFloatProperty("image.scale",31.429);
  }
  catch(IOException e) {
    println("couldn't read config file...");
  }

  try{
    p = loadImage("settings/background.jpg");
    size(p.width+200, p.height);
    imageLoaded = true;
  }
  catch(NullPointerException e){
    print("\nUsing default size as no background image available\n\n");
    size(width,height);
  }
  
  fill(0, 0, 255, 150);
  font = loadFont("ArialMT-12.vlw"); 
  textFont(font); 
  textAlign(LEFT);
  smooth();
  background(255, 255, 255); 
  stroke(0, 0, 0);
  strokeWeight(1);
  plants = new ArrayList();
  sfruit = new ArrayList();
  drawnPlants = new ArrayList();
  readData();  
  drawScale();
  drawImage();
  fill(0, 255, 0, 50);
  drawPlants(); 
  drawSFNames();
  drawTNames();
  drawLegend();
}

void draw() {
  if (keyPressed == true) {
    if (key == 'g' || key == 'G') {
      drawGrid();
    }
    if (key == 'e' || key =='E') {
      if (!editMode) {
        editMode = true;
        print("\n\nEntering EDIT mode (x to leave)\n\n");
        //drawEditMode(editMode);
      }
    }
    if (key == 'v' || key =='V') {
      editMode = false;
      //drawEditMode(editMode);
    } 
    if (key == 'p') {
      editMode = false;
      background(255, 255, 255);  
      drawImage();
      drawScale(); 
      drawPlants();
      drawTNames();
      drawSFNames(); 
      drawLegend(); 
      readDrawPlotBoundary();
    }
    if (key == 'P') {
      editMode = false;
      background(255, 255, 255);  
      drawImage();
      drawScale(); 
      drawPlants();
      readDrawPlotBoundary();
    }
    if (key == 't') {
      editMode = false;
      background(255, 255, 255);   
      drawImage();  
      drawScale(); 
      drawTrees();
      drawTNames();
      readDrawPlotBoundary();
    }  
    if (key == 'T') {
      editMode = false;
      background(255, 255, 255);     
      drawImage();
      drawScale(); 
      drawTrees();
      readDrawPlotBoundary();
    }  

    if (key == 's') {
      editMode = false;
      background(255, 255, 255); 
      drawImage();
      drawScale();
      drawSFruit();    
      drawSFNames(); 
      drawLegend(); 
      readDrawPlotBoundary();
    }
    if (key == 'S') {
      editMode = false;
      background(255, 255, 255); 
      drawImage();
      drawScale();
      drawSFruit();    
      drawLegend(); 
      readDrawPlotBoundary();
    }

    if (key == 'n' || key == 'N') {
      editMode = false;
      drawTNames();
    }  
    if (key == 'b' || key == 'b') {
      imageOn = true;
      editMode = false;
      background(255, 255, 255);  
      drawImage();
      drawScale(); 
      drawPlants();
      drawTNames();
      drawSFNames(); 
      drawLegend(); 
      readDrawPlotBoundary();
    }  
    if (key == 'x' || key == 'X') {
      imageOn = false;
      editMode = false;
      background(255, 255, 255);  
      drawImage();
      drawScale(); 
      drawPlants();
      drawTNames();
      drawSFNames(); 
      drawLegend(); 
      readDrawPlotBoundary();
    } 
    if (key == 'w' || key == 'W') {
      writeDrawnPlants();
    } 
    if (key == 'd' || key == 'D') {
      //    editMode = !editMode;
      //       datesOn = !datesOn;
      //       background(255,255,255);  
      //       drawScale(); 
      //       drawPlants();
      //       readDrawPlotBoundary();  
      //       drawTNames();
    }
    if (key == 'i' || key == 'I') {
      save("settings/plan.png");
      print("\nSaved plan.png\n");
    }
    if (key == 'h' || key == 'H') {
      //if (frameCount<(frameTrig+10)) {
        showInstr = false;
        editMode = false;
        background(255, 255, 255);  
        drawImage();
        drawScale(); 
        drawPlants();
        drawTNames();
        drawSFNames(); 
        drawLegend(); 
        readDrawPlotBoundary();
       // frameTrig = frameCount;
      //}
    }  
    if (key == 'j' || key == 'J') {
        showInstr = true;
        editMode = false;
        background(255, 255, 255);  
        drawImage();
        drawScale(); 
        drawPlants();
        drawTNames();
        drawSFNames(); 
        drawLegend(); 
        readDrawPlotBoundary();
    }
  }
}

void mousePressed() {
  if (editMode) {
    centerX = mouseX;
    centerY = mouseY;
  }
}
void mouseDragged() {
  if (editMode) {
    outerX = mouseX;
    outerY = mouseY;
    float x = outerX - centerX;
    float y = outerY - centerY;
    float diam = 2*sqrt(x*x + y*y);
    stroke(0, 0, 255, 50);
    fill(0, 0, 0, 0);
    ellipse(centerX, centerY, diam, diam);
  }
}
void mouseReleased() {
  if (editMode) {
    outerX = mouseX;
    outerY = mouseY;
    float xd = outerX - centerX;
    float yd = outerY - centerY;
    float diam = 2*sqrt(xd*xd + yd*yd);
    stroke(0, 255, 0, 150);
    fill(0, 255, 0, 150);
    ellipse(centerX, centerY, diam, diam);
    String b = "no data";
    float r = diam/scale;
    float x = centerX/scale;
    float y = (centerY+offset-height)/(-scale);
    
    drawnPlants.add(new Plant(x, y, r, b, b, b, false));
    print("\n\nNew plant with radius: "+r+" created at: "+x+","+y+"\nPress w to write locations to data/drawn.csv file\n\n");
    //new Plant(x,y,r,b,b,b,b,b,b,false);
  }
}
void mouseMoved() {
  if (editMode) {
 float x = mouseX/scale;
 float y = (mouseY+offset-height)/(-scale); 
 print("EDIT mode (x to leave), pixels from top right:("+mouseX+","+mouseY+") scaled location from bottom left:("+x+","+y+")\n"); 
 }
}

class Plant {
  float x;
  float y;
  float d;
  String species;
  String variety; 
  String rootstock;
  String plantingDate;
  String notes;
  String code;
  String legEntry;
  boolean legPrint;
  Plant(float xin, float yin, float din, String sp, String var, String cd, boolean LP) {
    x = xin;
    y = yin;
    d = din;
    species = sp;
    variety = var;
    code = cd;
    legPrint = LP;
  }
  float drawLegend( float x, float y, float os) 
  {
    fill(0, 0, 0);
    textAlign(LEFT);
    if (legPrint) {
      //legEntry = code + " \t \t " + species;  
      //text(legEntry,x,y);
      text(code, x+215, y);
      text(species, x+240, y);
      return y + os;
    }  
    return y;
  }
  void drawGlyph()
  {
    stroke(0, 255, 0, 150);
    ellipse(x*scale, height-y*scale-offset, d*scale, d*scale);
    fill(125, 125, 0);
    ellipse(x*scale, height-y*scale-offset, 7, 7);
  }
  void drawSFGlyph()
  {
    stroke(0, 255, 0, 150);
    ellipse(x*scale, height-y*scale-offset, d*scale, d*scale);
    fill(125, 125, 0);
    //ellipse(x*scale, height-y*scale-offset, 7, 7);
  }
  void drawDPGlyph()
  {
    stroke(0, 0, 255, 150);
    ellipse(x*scale, height-y*scale-offset, d*scale, d*scale);
    fill(125, 125, 0);
    ellipse(x*scale, height-y*scale-offset, 7, 7);
  }
  void drawText(boolean datesOn, float rotAngle) {
    textAlign(LEFT);
    fill(0, 0, 0);
    if (datesOn) {
      text(plantingDate, x*scale, height-y*scale-offset-3.5*t_offs);
    }
    rotate(rotAngle);
    text(species, x*scale, height-y*scale-offset-2.0*t_offs);
    text(variety, x*scale, height-y*scale-offset-0.5*t_offs);
    rotate(-rotAngle);
  }
  void drawLineText(boolean datesOn) {
    textAlign(CENTER);
    fill(0, 0, 0);
    float ll = 0;
    float x2 = x*scale;
    float y2 = height-y*scale-offset;
    float xc = width/2;
    float yc = height/2;
    float hyp; 
    hyp = sqrt( pow((x2-xc), 2) + pow((y2-yc), 2));
    float xe = x2+(ll/hyp)*(x2-xc);
    float ye = y2+(ll/hyp)*(y2-yc);

    if (datesOn) {
      text(plantingDate, x*scale, height-y*scale-offset-3.5*t_offs);
    }

    text(code, xe, ye+0.5*t_offs);
    //text(variety, xe, ye-0.5*t_offs);

    line(x2, y2, xe, ye);
  }
}

 
/**
configfiles taken from http://wiki.processing.org/index.php/External_configuration_files
@author toxi
*/
class P5Properties extends Properties {
 
  boolean getBooleanProperty(String id, boolean defState) {
    return boolean(getProperty(id,""+defState));
  }
 
  int getIntProperty(String id, int defVal) {
    return int(getProperty(id,""+defVal)); 
  }
 
  float getFloatProperty(String id, float defVal) {
    return float(getProperty(id,""+defVal)); 
  }  
}

